@@ -5,7 +5,10 @@ use Moose;
 use Scalar::Util qw(refaddr);
 
 use Sub::Exporter -setup => {
-    exports => [ map { $_ => \&_build_export } qw(fold unfold flatten unflatten) ],
+    exports => [
+        (map { $_ => \&_build_unary_export } qw(fold unfold flatten unflatten)),
+        (map { $_ => \&_build_nary_export  } qw(merge)),
+    ],
 };
 
 use constant {
@@ -15,7 +18,7 @@ use constant {
     VALUE => 1,
 };
 
-our $VERSION = '0.0.2';
+our $VERSION = '0.1.0';
 
 has on_object => (
     isa      => 'CodeRef',
@@ -23,7 +26,7 @@ has on_object => (
     default  => sub { sub { $_[1] } }, # return the value unchanged
 );
 
-has on_seen => (
+has on_cycle => (
     isa      => 'CodeRef',
     is       => 'ro',
     default  => sub { sub { } }, # do nothing
@@ -41,44 +44,45 @@ has array_delimiter => (
     default  => '.',
 );
 
-sub fold {
-    my ($self, $value) = @_;
-    my $ref = ref($value);
+around BUILDARGS => sub {
+    my $original = shift;
+    my $class = shift;
+    my $args = $class->$original(@_);
+    my $delimiter = delete $args->{delimiter};
 
-    if ($ref eq 'HASH') {
-        my $prefix = undef;
-        my $target = {};
-        my $seen = {};
-
-        my $hash = $self->_merge($value, $prefix, $target, $seen);
-        return $hash;
-    } else {
-        my $type = length($ref) ? "'$ref'" : 'non-reference';
-        confess "invalid argument: expected unblessed HASH reference, got: $type";
+    if (defined $delimiter) {
+        $args->{array_delimiter} = $delimiter unless (exists $args->{array_delimiter});
+        $args->{hash_delimiter}  = $delimiter unless (exists $args->{hash_delimiter});
     }
+
+    return $args;
+};
+
+sub fold {
+    my $self = shift;
+    my $hash = _check_hash(shift);
+    my $prefix = undef;
+    my $target = {};
+    my $seen = {};
+
+    return $self->_merge($hash, $prefix, $target, $seen);
 }
 
 sub unfold {
-    my ($self, $hash) = @_;
-    my $ref = ref($hash);
+    my $self = shift;
+    my $hash = _check_hash(shift);
+    my $target = {};
 
-    if ($ref eq 'HASH') {
-        my $target = {};
-
-        # sorting the keys should lead to better locality of reference,
-        # for what that's worth here
-        # XXX the sort order is connected with the ambiguity issue mentioned below
-        for my $key (sort keys %$hash) {
-            my $value = $hash->{$key};
-            my $steps = $self->_split($key);
-            $self->_set($target, $steps, $value);
-        }
-
-        return $target;
-    } else {
-        my $type = length($ref) ? "'$ref'" : 'non-reference';
-        confess "invalid argument: expected unblessed HASH reference, got: $type";
+    # sorting the keys should lead to better locality of reference,
+    # for what that's worth here
+    # XXX the sort order is connected with the ambiguity issue mentioned below
+    for my $key (sort keys %$hash) {
+        my $value = $hash->{$key};
+        my $steps = $self->_split($key);
+        $self->_set($target, $steps, $value);
     }
+
+    return $target;
 }
 
 BEGIN {
@@ -86,23 +90,66 @@ BEGIN {
     *unflatten = \&unfold;
 }
 
+sub merge {
+    my $self = shift;
+    return $self->unfold({ map { %{ $self->fold(_check_hash($_)) } } @_ });
+}
+
 sub is_object {
     my ($self, $value) = @_;
     my $ref = ref($value);
     return $ref && ($ref ne 'HASH') && ($ref ne 'ARRAY');
 }
 
-sub _build_export {
+sub _build_unary_export {
     my ($class, $name, $base_options) = @_;
 
-    return sub ($;@) {
-        my $hash = shift;
+    return sub {
+        my $arg = shift;
         my $custom_options = @_ == 1 ? shift : { @_ };
         my $folder = $class->new({ %$base_options, %$custom_options });
-        return $folder->$name($hash);
+        return $folder->$name($arg);
     }
 }
 
+sub _build_nary_export {
+    my ($class, $name, $base_options) = @_;
+
+    return sub {
+        my ($args, $custom_options);
+
+        if (@_ && (ref($_[0]) eq 'ARRAY')) {
+            $args = shift;
+            $custom_options = @_ == 1 ? shift : { @_ };
+        } else {
+            $args = [ @_ ];
+            $custom_options = {};
+        }
+
+        my $folder = $class->new({ %$base_options, %$custom_options });
+        return $folder->$name(@$args);
+    }
+}
+
+sub _check_hash {
+    my $hash = shift;
+    my $ref = ref($hash);
+
+    unless ($ref eq 'HASH') {
+        my $type;
+
+        if (defined $hash) {
+            $type = length($ref) ? "'$ref'" : 'non-reference';
+        } else {
+            $type = 'undef';
+        }
+
+        confess "invalid argument: expected unblessed HASH reference, got: $type";
+    }
+
+    return $hash;
+}
+
 sub _join {
     my ($self, $prefix, $delimiter, $key) = @_;
     return defined($prefix) ? $prefix . $delimiter . $key : $key;
@@ -129,7 +176,7 @@ The ambiguity can either be resolved here/in unfold with a bit of static analysi
 lazily/dynamically in _set (need to sort the keys so that non-integers (if any) are unpacked
 before integers (if any)).
 
-Currently, the example above is unpacked corectly :-)
+Currently, the example above is unpacked correctly :-)
 
 =end comment
 
@@ -200,28 +247,36 @@ sub _merge {
         #
         # if the callback doesn't raise a fatal exception,
         # treat the value as a terminal
-        $self->on_seen->($self, $value); # might warn or die
+        $self->on_cycle->($self, $value); # might warn or die
         $target->{$target_key} = $value; # treat as a terminal
     } elsif ($ref eq 'HASH') {
         my $delimiter = $self->hash_delimiter;
 
         $seen->{$refaddr} = 1;
 
-        # sorting the keys ensures a deterministic order,
-        # which (at the very least) is required for unsurprising
-        # tests
-        for my $hash_key (sort keys %$value) {
-            my $hash_value = $value->{$hash_key};
-            $self->_merge($hash_value, $self->_join($target_key, $delimiter, $hash_key), $target, $seen);
+        if (%$value) {
+            # sorting the keys ensures a deterministic order,
+            # which (at the very least) is required for unsurprising
+            # tests
+            for my $hash_key (sort keys %$value) {
+                my $hash_value = $value->{$hash_key};
+                $self->_merge($hash_value, $self->_join($target_key, $delimiter, $hash_key), $target, $seen);
+            }
+        } else {
+            $target->{$target_key} = {};
         }
     } elsif ($ref eq 'ARRAY') {
         my $delimiter = $self->array_delimiter;
 
         $seen->{$refaddr} = 1;
 
-        for my $index (0 .. $#$value) {
-            my $array_element = $value->[$index];
-            $self->_merge($array_element, $self->_join($target_key, $delimiter, $index), $target, $seen);
+        if (@$value) {
+            for my $index (0 .. $#$value) {
+                my $array_element = $value->[$index];
+                $self->_merge($array_element, $self->_join($target_key, $delimiter, $index), $target, $seen);
+            }
+        } else {
+            $target->{$target_key} = [];
         }
     } else { # terminal
         $target->{$target_key} = $value;
@@ -285,6 +340,7 @@ Hash::Fold - flatten and unflatten nested hashrefs
     use Hash::Fold qw(flatten unflatten);
 
     my $object = bless { foo => 'bar' };
+
     my $nested = {
         foo => $object,
         baz => {
@@ -294,6 +350,7 @@ Hash::Fold - flatten and unflatten nested hashrefs
     };
 
     my $flattened = flatten($nested);
+
     is_deeply $flattened, {
         'baz.a'     => 'b',
         'baz.c.0'   => 'd',
@@ -303,39 +360,159 @@ Hash::Fold - flatten and unflatten nested hashrefs
     };
 
     my $roundtrip = unflatten($flattened);
+
     is_deeply $roundtrip, $nested;
 
 =head1 DESCRIPTION
 
-This module provides functional and OO interfaces which allow hashrefs containing nested
-values to be flattened into single-level hashrefs (e.g. with dotted keys) and unflattened.
+This module provides functional and OO interfaces which can be used to flatten, unflatten
+and merge nested hashrefs.
+
+Unless noted, the functions listed below are also available as methods. Options provided to the
+C<Hash::Fold> constructor can be supplied to the functions e.g.:
+
+    use Hash::Fold;
+
+    my $folder = Hash::Fold->new(delimiter => '/');
+
+    $folder->fold($hash);
+
+is equivalent to:
+
+    use Hash::Fold qw(fold);
+
+    my $folded = fold($hash, delimiter => '/');
+
+Options (and constructor args) can be supplied as a list of pairs or a hashref, so the
+following are equivalent:
+
+    my $folded = fold($hash,   delimiter => '/'  );
+    my $folded = fold($hash, { delimiter => '/' });
+
+In addition, C<Hash::Fold> uses L<Sub::Exporter>, which allows functions to bake
+in options on import e.g.:
+
+    use Hash::Fold fold => { delimiter => '/' };
+
+    my $folded = fold($hash);
+
+=head1 OPTIONS
+
+As described above, the following options can be supplied as constructor args, import args,
+or per-function overrides. Under the hood, they are (L<Moose>) attributes which can be wrapped
+and overridden like any other attributes.
+
+=head2 array_delimiter (Str, ro)
+
+The delimiter prefixed to array elements when flattening and unflattening. Default: C<.>.
+
+=head2 hash_delimiter (Str, ro)
+
+The delimiter prefixed to hash elements when flattening and unflattening. Default: C<.>.
+
+=head2 delimiter (Str)
+
+This is effectively a write-only attribute which assigns the same string to L<"array_delimiter"> and
+L<"hash_delimiter">. It can only be supplied as a constructor arg or function option (which are
+equivalent) i.e. C<Hash::Fold> instances have no C<delimiter> method.
+
+=head2 on_cycle (CodeRef, ro)
+
+A callback invoked whenever L<"fold"> encounters a circular reference i.e. a reference which contains
+itself as a nested value.
+
+The callback takes two arguments: the C<Hash::Fold> instance and the value e.g.:
+
+    sub on_cycle {
+        my ($folder, $value) = @_;
+        warn 'self-reference found: ', Dumper(value), $/;
+    }
+
+    my $folder = Hash::Fold->new(on_cycle => \&on_cycle);
+
+Note that circular references are handled correctly i.e. they are treated as terminals and not traversed.
+This callback merely provides a mechanism to report them (e.g. by issuing a warning).
+
+The default callback does nothing.
+
+=head2 on_object (CodeRef, ro)
+
+A callback invoked whenever L<"fold"> encounters a value for which the L<"is_object"> method returns true
+i.e. any reference that isn't an unblessed arrayref or unblessed hashref. This callback can be used to modify
+the value e.g. to return a traversable value (e.g. unblessed hashref) in place of a terminal (e.g.
+blessed hashref).
+
+The callback takes two arguments: the C<Hash::Fold> instance and the object e.g.:
+
+    use Scalar::Util qw(blessed);
+
+    sub on_object {
+        my ($folder, $value) = @_;
+
+        if (blessed($value) && $value->isa('HASH')) {
+            return { %$value }; # unbless
+        } else {
+            return $value;
+        }
+    }
+
+    my $folder = Hash::Fold->new(on_object => \&on_object);
+
+The default callback returns its value unchanged.
 
 =head1 EXPORTS
 
 Nothing by default. The following functions can be imported.
 
-=head2 flatten
-
-Takes a nested hashref and returns a single-level hashref with (by default) dotted keys.
-
-Unblessed arrayrefs and unblessed hashrefs are traversed. All other values
-(e.g. strings, regexps, objects &c.) are treated as terminals and passed through verbatim.
-
 =head2 fold
 
-Provided as an alias for L<"flatten">.
+Takes a nested hashref and returns a single-level hashref with (by default) dotted keys.
+The delimiter can be overridden via the L<"delimiter">, L<"array_delimiter"> and
+L<"hash_delimiter"> options.
 
-=head2 unflatten
+Unblessed arrayrefs and unblessed hashrefs are traversed. All other values (e.g. strings,
+regexps, objects &c.) are treated as terminals and passed through verbatim,
+although this can be overridden by supplying a suitable L<"on_object"> callback.
 
-Takes a flattened hashref and returns the corresponding nested hashref.
+=head2 flatten
+
+Provided as an alias for L<"fold">.
 
 =head2 unfold
 
-Provided as an alias for L<"unflatten">.
+Takes a flattened hashref and returns the corresponding nested hashref.
+
+=head2 unflatten
+
+Provided as an alias for L<"unfold">.
+
+=head2 merge
+
+Takes a list of hashrefs which are then flattened, merged into one (in the order provided i.e.
+with precedence given to the rightmost arguments) and unflattened i.e. shorthand for:
+
+    unflatten { map { %{ flatten $_ } } @_ }
+
+To provide options to the C<merge> subroutine, pass the hashrefs in an arrayref, and the options
+(as usual) as a list of pairs or a hashref:
+
+    merge([ $hash1, $hash2, ... ],   delimiter => ...  )
+    merge([ $hash1, $hash2, ... ], { delimiter => ... })
+
+=head1 METHODS
+
+=head2 is_object
+
+This method is called from L<"fold"> to determine whether a value should be passed to the L<"on_object">
+callback.
+
+It is passed each value encountered while traversing a hashref and returns true for all references (e.g.
+regexps, globs, objects &c.) apart from unblessed arrayrefs and unblessed hashrefs, and false for all other
+values (i.e. unblessed hashrefs, unblessed arrayrefs, and non-references).
 
 =head1 VERSION
 
-0.0.2
+0.1.0
 
 =head1 SEE ALSO
 
@@ -345,6 +522,10 @@ Provided as an alias for L<"unflatten">.
 
 =item * L<Hash::Flatten>
 
+=item * L<Hash::Merge>
+
+=item * L<Hash::Merge::Simple>
+
 =back
 
 =head1 AUTHOR
